/*

Copyright (C) 2001-2002       A Nourai

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the included (GNU.txt) GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "quakedef.h"
#include "crc.h"
#include "gl_local.h"
#include "gl_image.h"

qboolean OnChange_gl_max_size (cvar_t *var, char *string);
qboolean OnChange_gl_texturemode (cvar_t *var, char *string);

char *COM_FileExtension (char *in);

static qboolean no24bit, forceTextureReload;

extern unsigned d_8to24table2[256];	//d_8to24table is in alpha-b-g-r format
extern byte vid_gamma_table[256];
extern float vid_gamma;

int texture_extension_number = 1;
int	gl_max_size_default;				
int	gl_lightmap_format = 3, gl_solid_format = 3, gl_alpha_format = 4;

cvar_t	gl_max_size		= {"gl_max_size", "1024", 0, OnChange_gl_max_size};
cvar_t	gl_picmip		= {"gl_picmip", "0"};
cvar_t	gl_lerpimages	= {"gl_lerpimages", "1"};
cvar_t	gl_texturemode	= {"gl_texturemode", "GL_LINEAR_MIPMAP_NEAREST", 0, OnChange_gl_texturemode};
cvar_t	gl_picmip_all	= {"gl_picmip_all", "1"};

typedef struct {
	int			texnum;
	char		identifier[MAX_QPATH];
	char		*pathname;	
	int			width, height;
	int			scaled_width, scaled_height;
	char		mipmap;
	qboolean	brighten;
	unsigned	crc;
	int			bpp;	
} gltexture_t;

gltexture_t	gltextures[MAX_GLTEXTURES];
int	numgltextures;


qboolean OnChange_gl_max_size (cvar_t *var, char *string) {
	int i;
	float newvalue = Q_atof(string);

	if (newvalue > gl_max_size_default) {
		Com_Printf("Your hardware doesn't support texture sizes bigger than %dx%d\n", gl_max_size_default, gl_max_size_default);
		return true;
	}

	for (i = 1; i < newvalue; i <<= 1)
		;

	if (i != newvalue) {
		Com_Printf("Incorrect value for %s: %d is not a power of 2\n", var->name, (int) newvalue);
		return true;
	}

	return false;
}

typedef struct {
	char *name;
	int	minimize, maximize;
} glmode_t;

glmode_t modes[] = {
	{"GL_NEAREST", GL_NEAREST, GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR}
};

#define GLMODE_NUMODES	(sizeof(modes) / sizeof(glmode_t))

static int gl_filter_min = GL_LINEAR_MIPMAP_NEAREST;
static int gl_filter_max = GL_LINEAR;

qboolean OnChange_gl_texturemode (cvar_t *var, char *string) {
	int i;
	gltexture_t	*glt;

	for (i = 0; i < GLMODE_NUMODES; i++) {
		if (!Q_strcasecmp (modes[i].name, string ) )
			break;
	}
	if (i == GLMODE_NUMODES) {
		Com_Printf ("bad filter name: %s\n", string);
		return true;
	}

	gl_filter_min = modes[i].minimize;
	gl_filter_max = modes[i].maximize;

	// change all the existing mipmap texture objects
	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++) {
		if (glt->mipmap) {
			GL_Bind (glt->texnum);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
		}
	}
	return false;
}


static void ResampleTextureLerpLine (byte *in, byte *out, int inwidth, int outwidth) {
	int j, xi, oldx = 0, f, fstep, endx;

	fstep = (int) (inwidth * 65536.0f / outwidth);
	endx = (inwidth - 1);
	for (j = 0, f = 0; j < outwidth; j++, f += fstep) {
		xi = (int) f >> 16;
		if (xi != oldx) {
			in += (xi - oldx) * 4;
			oldx = xi;
		}
		if (xi < endx) {
			int lerp = f & 0xFFFF;
			*out++ = (byte) ((((in[4] - in[0]) * lerp) >> 16) + in[0]);
			*out++ = (byte) ((((in[5] - in[1]) * lerp) >> 16) + in[1]);
			*out++ = (byte) ((((in[6] - in[2]) * lerp) >> 16) + in[2]);
			*out++ = (byte) ((((in[7] - in[3]) * lerp) >> 16) + in[3]);
		} else  {		// last pixel of the line has no pixel to lerp to
			*out++ = in[0];
			*out++ = in[1];
			*out++ = in[2];
			*out++ = in[3];
		}
	}
}

static void ResampleTexture (unsigned *indata, int inwidth, int inheight, unsigned *outdata, int outwidth, int outheight) {
	if (gl_lerpimages.value) {
		int i, j, yi, oldy, f, fstep, endy = (inheight - 1);
		byte *inrow, *out, *row1, *row2, *memalloc;

		out = (byte *) outdata;
		fstep = (int) (inheight * 65536.0f / outheight);

		memalloc = Q_Malloc(2 * (outwidth * 4));	
		row1 = memalloc;
		row2 = memalloc + (outwidth * 4);
		inrow = (byte *) indata;
		oldy = 0;
		ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
		ResampleTextureLerpLine (inrow + inwidth*4, row2, inwidth, outwidth);
		for (i = 0, f = 0; i < outheight; i++, f += fstep)	{
			yi = f >> 16;
			if (yi < endy) {
				int lerp = f & 0xFFFF;
				if (yi != oldy) {
					inrow = (byte *)indata + inwidth * 4 * yi;
					if (yi == oldy + 1)
						memcpy(row1, row2, outwidth*4);
					else
						ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
					ResampleTextureLerpLine (inrow + inwidth * 4, row2, inwidth, outwidth);
					oldy = yi;
				}
				for (j = outwidth ; j ; j--) {
					out[0] = (byte) ((((row2[0] - row1[0]) * lerp) >> 16) + row1[0]);
					out[1] = (byte) ((((row2[1] - row1[1]) * lerp) >> 16) + row1[1]);
					out[2] = (byte) ((((row2[2] - row1[2]) * lerp) >> 16) + row1[2]);
					out[3] = (byte) ((((row2[3] - row1[3]) * lerp) >> 16) + row1[3]);
					out += 4;
					row1 += 4;
					row2 += 4;
				}
				row1 -= outwidth * 4;
				row2 -= outwidth * 4;
			} else {
				if (yi != oldy) {
					inrow = (byte *) indata + inwidth * 4 * yi;
					if (yi == oldy+1)
						memcpy(row1, row2, outwidth * 4);
					else
						ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
					oldy = yi;
				}
				memcpy(out, row1, outwidth * 4);
			}
		}
		free(memalloc);
	} else {
		int i, j;
		unsigned int frac, fracstep, *inrow, *out;

		out = outdata;

		fracstep = inwidth * 0x10000 / outwidth;
		for (i = 0; i < outheight; i++) {
			inrow = (int *) indata + inwidth * (i * inheight / outheight);
			frac = fracstep >> 1;
			for (j = outwidth >> 2 ; j ; j--) {
				out[0] = inrow[frac >> 16]; frac += fracstep;
				out[1] = inrow[frac >> 16]; frac += fracstep;
				out[2] = inrow[frac >> 16]; frac += fracstep;
				out[3] = inrow[frac >> 16]; frac += fracstep;
				out += 4;
			}
			for (j = outwidth & 3; j; j--) {
				*out = inrow[frac >> 16]; frac += fracstep;
				out++;
			}
		}
	}
}

static void MipMap (byte *in, int width, int height) {
	int i, j;
	byte *out;

	width <<= 2;
	height >>= 1;
	out = in;
	for (i = 0; i < height; i++, in += width) {
		for (j = 0; j < width; j += 8, out += 4, in += 8) {
			out[0] = (in[0] + in[4] + in[width + 0] + in[width + 4]) >> 2;
			out[1] = (in[1] + in[5] + in[width + 1] + in[width + 5]) >> 2;
			out[2] = (in[2] + in[6] + in[width + 2] + in[width + 6]) >> 2;
			out[3] = (in[3] + in[7] + in[width + 3] + in[width + 7]) >> 2;
		}
	}
}


int currenttexture = -1;				// to avoid unnecessary texture sets

void GL_Bind (int texnum) {
	if (currenttexture == texnum)
		return;

	currenttexture = texnum;
	glBindTexture (GL_TEXTURE_2D, texnum);
}

static GLenum oldtarget = GL_TEXTURE0_ARB;
static int cnttextures[4] = {-1, -1, -1, -1};     // cached
static qboolean mtexenabled = false;


void GL_SelectTexture (GLenum target) {
	if (target == oldtarget) 
		return;

	qglActiveTexture (target);

	cnttextures[oldtarget - GL_TEXTURE0_ARB] = currenttexture;
	currenttexture = cnttextures[target - GL_TEXTURE0_ARB];
	oldtarget = target;
}

void GL_DisableMultitexture (void) {
	if (mtexenabled) {
		glDisable (GL_TEXTURE_2D);
		GL_SelectTexture (GL_TEXTURE0_ARB);
		mtexenabled = false;
	}
}

void GL_EnableMultitexture (void) {
	if (gl_mtexable) {
		GL_SelectTexture (GL_TEXTURE1_ARB);
		glEnable (GL_TEXTURE_2D);
		mtexenabled = true;
	}
}

void GL_EnableTMU (GLenum target) {
	GL_SelectTexture(target);
	glEnable(GL_TEXTURE_2D);
}

void GL_DisableTMU(GLenum target) {
	GL_SelectTexture(target);
	glDisable(GL_TEXTURE_2D);
}

static void ScaleDimensions(qboolean mipmap, int width, int height, int *scaled_width, int *scaled_height) {
	int maxsize;

	for (*scaled_width = 1; *scaled_width < width; *scaled_width <<= 1)
		;
	for (*scaled_height = 1; *scaled_height < height; *scaled_height <<= 1)
		;

	if (mipmap) {
		*scaled_width >>= (int) gl_picmip.value;
		*scaled_height >>= (int) gl_picmip.value;		
	}

	
	maxsize = mipmap ? gl_max_size.value : gl_max_size_default;

	*scaled_width = bound(1, *scaled_width, maxsize);
	*scaled_height = bound(1, *scaled_height, maxsize);
}

void GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap, qboolean alpha) {
	int	internal_format, scaled_width, scaled_height, miplevel;
	unsigned int *scaled;

	ScaleDimensions(mipmap, width, height, &scaled_width, &scaled_height);

	scaled = Q_Malloc(scaled_width * scaled_height * 4);

	internal_format = alpha ? gl_alpha_format : gl_solid_format;

	if (mipmap) {
		if (scaled_width == width && scaled_height == height)
			memcpy (scaled, data, width * height * 4);
		else
			ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);

		glTexImage2D (GL_TEXTURE_2D, 0, internal_format, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
		miplevel = 0;
		while (scaled_width > 1 || scaled_height > 1) {
			MipMap ((byte *) scaled, scaled_width, scaled_height);
			scaled_width >>= 1;
			scaled_height >>= 1;
			scaled_width = max(1, scaled_width);
			scaled_height = max(1, scaled_height);
			miplevel++;
			glTexImage2D (GL_TEXTURE_2D, miplevel, internal_format, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
		}
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	} else {
		if (scaled_width == width && scaled_height == height) {
			glTexImage2D (GL_TEXTURE_2D, 0, internal_format, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
		} else {
			ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);
			glTexImage2D (GL_TEXTURE_2D, 0, internal_format, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
		}
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_max);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}

	free(scaled);
}

void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean alpha, qboolean brighten) {
	static unsigned trans[640 * 480];
	int	i, image_size, p;
	unsigned *table;

	table = brighten ? d_8to24table2 : d_8to24table;
	image_size = width * height;

	if (image_size * 4 > sizeof(trans))
		Sys_Error ("GL_Upload8: image too big");

	if (alpha == 2)	{	//fullbright
		for (i = 0; i < image_size; i++) {
			p = data[i];
			if (p < 224)
				trans[i] = table[p] & 0x00FFFFFF;	// transparent 
			else
				trans[i] = table[p];				// fullbright
		}
	} else if (alpha) {
		// if there are no transparent pixels, make it a 4 component
		// texture even if it was specified as otherwise
		alpha = false;
		for (i = 0; i < image_size; i++) {
			if ((p = data[i]) == 255)
				alpha = true;
			trans[i] = table[p];
		}
	} else {
		if (image_size & 3)
			Sys_Error ("GL_Upload8: image_size & 3");
		for (i = 0; i < image_size; i += 4) {
			trans[i] = table[data[i]];
			trans[i + 1] = table[data[i + 1]];
			trans[i + 2] = table[data[i + 2]];
			trans[i + 3] = table[data[i + 3]];
		}
	}
	GL_Upload32 (trans, width, height, mipmap, alpha);
}

int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, qboolean brighten, int bpp) {
	int	i, scaled_width, scaled_height, crc = 0;
	gltexture_t	*glt;

	if (lightmode != 2)
		brighten = false;

	ScaleDimensions(mipmap, width, height, &scaled_width, &scaled_height);

	// see if the texture is already present
	if (identifier[0]) {
		crc = CRC_Block (data, width * height * bpp);
		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++) {
			if (!strncmp (identifier, glt->identifier, sizeof(glt->identifier) - 1)) {
				if (
					width == glt->width && height == glt->height &&
					scaled_width == glt->scaled_width && scaled_height == glt->scaled_height &&
					crc == glt->crc && brighten == glt->brighten && glt->bpp == bpp && glt->mipmap == mipmap
				) {
					GL_Bind(gltextures[i].texnum);
					return gltextures[i].texnum;	//texture cached
				} else {
					goto setup_gltexture;		// reload the texture into the same slot
				}
			}
		}
	}

	if (numgltextures == MAX_GLTEXTURES)
		Sys_Error ("GL_LoadTexture: numgltextures == MAX_GLTEXTURES");

	glt = &gltextures[numgltextures];
	numgltextures++;

	Q_strncpyz (glt->identifier, identifier, sizeof(glt->identifier));
	glt->texnum = texture_extension_number;
	texture_extension_number++;

setup_gltexture:
	glt->width = width;
	glt->height = height;
	glt->scaled_width = scaled_width;
	glt->scaled_height = scaled_height;
	glt->mipmap = mipmap;
	glt->brighten = brighten;
	glt->crc = crc;
	glt->bpp = bpp;
	if (glt->pathname) {
		Z_Free(glt->pathname);
		glt->pathname = NULL;
	}
	
	if (bpp == 4 && com_netpath[0])	
		glt->pathname = CopyString(com_netpath);

	GL_Bind(glt->texnum);

	switch (bpp) {
	case 1:
		GL_Upload8 (data, width, height, mipmap, alpha, brighten); break;
	case 4:
		GL_Upload32 ((void *) data, width, height, mipmap, alpha); break;
	default:
		Sys_Error("GL_LoadTexture: unknown bpp\n"); break;
	}

	return glt->texnum;
}

int GL_LoadPicTexture (char *name, mpic_t *pic, byte *data) {
	int glwidth, glheight, i;
	char fullname[MAX_QPATH] = "pic:";
	byte *src, *dest, *buf;

	for (glwidth = 1 ; glwidth < pic->width ; glwidth <<= 1)
		;
	for (glheight = 1 ; glheight < pic->height ; glheight <<= 1)
		;

	Q_strncpyz (fullname + 4, name, sizeof(fullname) - 4);
	if (glwidth == pic->width && glheight == pic->height) {
		pic->texnum = GL_LoadTexture (fullname, glwidth, glheight, data, false, true, false, 1);
		pic->sl = 0;
		pic->sh = 1;
		pic->tl = 0;
		pic->th = 1;
	} else {
		buf = Q_Calloc (glwidth * glheight, 1);

		src = data;
		dest = buf;
		for (i = 0; i < pic->height; i++) {
			memcpy (dest, src, pic->width);
			src += pic->width;
			dest += glwidth;
		}
		pic->texnum = GL_LoadTexture (fullname, glwidth, glheight, buf, false, true, false, 1);
		pic->sl = 0;
		pic->sh = (float) pic->width / glwidth;
		pic->tl = 0;
		pic->th = (float) pic->height / glheight;
		free (buf);
	}

	return pic->texnum;
}

gltexture_t *GL_FindTexture (char *identifier) {
	int i;

	for (i = 0; i < numgltextures; i++) {
		if (!strcmp (identifier, gltextures[i].identifier))
			return &gltextures[i];
	}

	return NULL;
}

static gltexture_t *current_texture = NULL;

#define CHECK_TEXTURE_ALREADY_LOADED	\
	if (CheckTextureLoaded(mode)) {		\
		current_texture = NULL;			\
		fclose(f);						\
		return NULL;					\
	}


static qboolean CheckTextureLoaded(int mode) {
	int scaled_width, scaled_height;
	qboolean mipmap;

	if (!forceTextureReload) {
		if (current_texture && current_texture->pathname && !strcmp(com_netpath, current_texture->pathname)) {
			
			mipmap = (mode & TEX_MIPMAP) ? true : false;
			ScaleDimensions(mipmap, current_texture->width, current_texture->height, &scaled_width, &scaled_height);
			if (current_texture->scaled_width == scaled_width && current_texture->scaled_height == scaled_height)
				return true;
		}
	}
	return false;
}


byte *GL_LoadImagePixels (char *filename, int matchwidth, int matchheight, int mode) {
	char basename[MAX_QPATH], name[MAX_QPATH];
	byte *c, *data;
	FILE *f;

	COM_StripExtension(filename, basename);
	for (c = basename; *c; c++)
		if (*c == '*')
			*c = '#';

	Q_snprintfz (name, sizeof(name), "%s.tga", basename);
	if (FS_FOpenFile (name, &f) != -1) {
		CHECK_TEXTURE_ALREADY_LOADED;
		if (data = Image_LoadTGA (f, name, matchwidth, matchheight))
			return data;
	}

#ifdef WITH_PNG
	Q_snprintfz (name, sizeof(name), "%s.png", basename);
	if (FS_FOpenFile (name, &f) != -1) {
		CHECK_TEXTURE_ALREADY_LOADED;
		if (data = Image_LoadPNG (f, name, matchwidth, matchheight))
			return data;
	}
#endif

	if (mode & TEX_COMPLAIN) {
		if (!no24bit)
			Com_Printf ("Couldn't load %s image\n", COM_SkipPath(filename));
	}

	return NULL;
}


int GL_LoadTexturePixels (byte *data, char *identifier, int width, int height, int mode) {
	int i, j, image_size;
	qboolean transparent = false, fb = false;

	if ((mode & (TEX_FULLBRIGHT|TEX_LUMA)) && !cl.userfb && !com_serveractive)	
		return 0;

	image_size = width * height;

	if (mode & TEX_LUMA) {
		transparent = true;
		for (i = 0; i < image_size; i++) { 
			if (!data[4 * i] && !data[4 * i + 1] && !data[4 * i + 2])
				data[4 * i + 3] = 0;
		}
	} else if (mode & TEX_FULLBRIGHT) {
		transparent = true;
		for (j = 0; j < image_size; j++) {
			for (i = 224; i < 256; i++) {
				if (d_8to24table[i] == ((unsigned *) data)[j]) {
					fb = true;
					break;
				}
			}
			if (i == 256)
				((unsigned *) data)[j] &= 0x00FFFFFF;
		}
		if (fb == false)
			return 0;
	} else if (mode & TEX_ALPHA) {
		for (j = 0; j < image_size; j++) {
			if ( ( (((unsigned *) data)[j] >> 24 ) & 0xFF ) < 255 ) {
				transparent = true;
				break;
			}
		}
	}

	if (vid_gamma != 1) {
		for (i = 0; i < image_size; i++){
			data[4 * i] = vid_gamma_table[data[4 * i]];
			data[4 * i + 1] = vid_gamma_table[data[4 * i + 1]];
			data[4 * i + 2] = vid_gamma_table[data[4 * i + 2]];
		}
	}

	return GL_LoadTexture (identifier, width, height, data, mode & TEX_MIPMAP, transparent, false, 4);
}


int GL_LoadTextureImage (char *filename, char *identifier, int matchwidth, int matchheight, int mode) {
	int texnum;
	byte *data;
	gltexture_t *gltexture;
	qboolean transparent = false, fb = false;

	if (no24bit)
		return 0;

	if ((mode & (TEX_FULLBRIGHT|TEX_LUMA)) && !cl.userfb && !com_serveractive)	
		return 0;

	if (!identifier)
		identifier = filename;

	gltexture = current_texture = GL_FindTexture(identifier);

	if (!(data = GL_LoadImagePixels (filename, matchwidth, matchheight, mode))) {
		
		
		texnum =  (gltexture && !current_texture) ? gltexture->texnum : 0;
	} else {
		texnum = GL_LoadTexturePixels(data, identifier, image_width, image_height, mode);
		free(data);
	}

	current_texture = NULL;
	return texnum;
}

mpic_t *GL_LoadPicImage (char *filename, char *id, int matchwidth, int matchheight, int mode) {
	int glwidth, glheight, i;
	char identifier[MAX_QPATH] = "pic:";
	byte *data, *src, *dest, *buf;
	qboolean transparent = false;
	static mpic_t pic;

	if (no24bit)
		return NULL;

	if (!(data = GL_LoadImagePixels (filename, matchwidth, matchheight, 0)))
		return NULL;

	pic.width = image_width;
	pic.height = image_height;

	if (mode & TEX_ALPHA) {
		for (i = 0; i < image_width * image_height; i++) {
			if ( ( (((unsigned *) data)[i] >> 24 ) & 0xFF ) < 255) {
				transparent = true;
				break;
			}
		}
	}

	for (glwidth = 1; glwidth < pic.width; glwidth <<= 1)
		;
	for (glheight = 1; glheight < pic.height; glheight <<= 1)
		;

	Q_strncpyz (identifier + 4, id ? id : filename, sizeof(identifier) - 4);
	if (glwidth == pic.width && glheight == pic.height) {
		pic.texnum = GL_LoadTexture (identifier, pic.width, pic.height, data, false, transparent, false, 4);
		pic.sl = 0;
		pic.sh = 1;
		pic.tl = 0;
		pic.th = 1;
	} else {
		buf = Q_Calloc (glwidth * glheight, 4);

		src = data;
		dest = buf;
		for (i = 0; i < pic.height; i++) {
			memcpy (dest, src, pic.width * 4);
			src += pic.width * 4;
			dest += glwidth * 4;
		}
		pic.texnum = GL_LoadTexture (identifier, glwidth, glheight, buf, false, transparent, false, 4);
		pic.sl = 0;
		pic.sh = (float) pic.width / glwidth;
		pic.tl = 0;
		pic.th = (float) pic.height / glheight;
		free (buf);
	}

	free(data);
	return &pic;
}

int GL_LoadCharsetImage (char *filename, char *identifier) {
	int i, j, texnum, image_size;
	byte *data, *buf, *dest, *src;
	qboolean transparent = false, fb = false;

	if (no24bit)
		return 0;

	if (!(data = GL_LoadImagePixels (filename, 0, 0, 0)))
		return 0;

	if (!identifier)
		identifier = filename;

	image_size = image_width * image_height;

	
	for (j = 0; j < image_size; j++) {
		if ( ( (((unsigned *) data)[j] >> 24 ) & 0xFF ) < 255 ) {
			transparent = true;
			break;
		}
	}
	if (!transparent) {
		for (i = 0 ; i < image_width * image_height; i++) {
			if ( data[i * 4] == data[0] && data[i * 4 + 1] == data[1] && data[i * 4 + 2] == data[2] )
				data[i * 4 + 3] = 0;
		}
	}

	
	buf = dest = Q_Calloc(image_size * 2, 4); 
	src = data;
	for (i = 0 ; i < 16 ; i++) {
		memcpy (dest, src, image_size >> 2);	
		src += image_size >> 2;
		dest += image_size >> 1;				
	}

	texnum = GL_LoadTexture (identifier, image_width, image_height * 2, buf, false, true, false, 4);

	free(buf);
	free(data);
	return texnum;
}

void GL_Texture_Init(void) {
	Cvar_SetCurrentGroup(CVAR_GROUP_TEXTURES);
	Cvar_Register(&gl_max_size);
	Cvar_Register(&gl_picmip);
	Cvar_Register(&gl_lerpimages);
	Cvar_Register(&gl_texturemode);
	Cvar_Register(&gl_picmip_all);

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &gl_max_size_default);
	Cvar_SetDefault(&gl_max_size, gl_max_size_default);	

	no24bit = COM_CheckParm("-no24bit") ? true : false;
	forceTextureReload = COM_CheckParm("-forceTextureReload") ? true : false;
}
